import tempfile

from GozargahNodeBridge.common.service_pb2 import User, Proxy, Vmess, Vless, Trojan, Shadowsocks


def string_to_temp_file(content: str):
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(content)
        return f.name


def create_user(email: str, proxies: Proxy, inbounds: list[str]) -> User:
    return User(email=email, proxies=proxies, inbounds=inbounds)


def create_proxy(
    vmess_id: str | None = None,
    vless_id: str | None = None,
    vless_flow: str | None = None,
    trojan_password: str | None = None,
    shadowsocks_password: str | None = None,
    shadowsocks_method: str | None = None,
) -> Proxy:
    return Proxy(
        vmess=Vmess(id=vmess_id),
        vless=Vless(id=vless_id, flow=vless_flow),
        trojan=Trojan(password=trojan_password),
        shadowsocks=Shadowsocks(password=shadowsocks_password, method=shadowsocks_method),
    )



import asyncio

from grpclib.client import Channel
from grpclib.config import Configuration
from grpclib.exceptions import GRPCError, StreamTerminatedError
from grpc import StatusCode as Status
from aiorwlock import RWLock

from GozargahNodeBridge.common import service_pb2 as service
from GozargahNodeBridge.common import service_grpc
from GozargahNodeBridge.controller import NodeAPIError, Health
from GozargahNodeBridge.abstract_node import GozargahNode


class Node(GozargahNode):
    def __init__(
        self,
        address: str,
        port: int,
        client_cert: str,
        client_key: str,
        server_ca: str,
        extra: dict | None = None,
        max_logs: int = 1000,
    ):
        super().__init__(client_cert, client_key, server_ca, extra, max_logs)
        try:
            self.channel = Channel(host=address, port=port, ssl=self.ctx, config=Configuration(_keepalive_timeout=10))
            self._client = service_grpc.NodeServiceStub(self.channel)
        except Exception as e:
            self._cleanup_temp_files()
            raise NodeAPIError(-1, f"Channel initialization failed: {str(e)}")

        self._metadata = None
        self._node_lock = RWLock()

    def _close_chan(self):
        """Close gRPC channel"""
        if hasattr(self, "channel"):
            try:
                self.channel.close()
            except Exception:
                pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()
        self._cleanup_temp_files()
        self._close_chan()

    def __del__(self):
        self._cleanup_temp_files()
        self._close_chan()

    async def _handle_error(self, error: Exception):
        """Convert gRPC errors to NodeAPIError."""
        if isinstance(error, asyncio.TimeoutError):
            raise NodeAPIError(-1, "Request timed out")

        elif isinstance(error, GRPCError):
            status_code = error.status
            if status_code == Status.UNAVAILABLE:
                raise NodeAPIError(-1, f"Node unavailable: {error.message}")
            else:
                raise NodeAPIError(status_code.value, error.message)

        elif isinstance(error, StreamTerminatedError):
            raise NodeAPIError(-2, f"Stream terminated: {str(error)}")
        else:
            raise NodeAPIError(0, str(error))

    async def _handle_grpc_request(self, method, request, timeout=15):
        """Handle a gRPC request and convert errors to NodeAPIError."""
        try:
            return await asyncio.wait_for(method(request, metadata=self._metadata), timeout=timeout)
        except Exception as e:
            await self._handle_error(e)

    async def start(
        self,
        config: str,
        backend_type: service.BackendType,
        users: list[service.User],
        keep_alive: int = 0,
        timeout: int = 15,
    ) -> service.BaseInfoResponse | None:
        """Start the node"""
        health = await self.get_health()
        if health in (Health.BROKEN, Health.HEALTHY):
            await self.stop()
        elif health is Health.INVALID:
            raise NodeAPIError(code=-4, detail="Invalid node")

        req = service.Backend(type=backend_type, config=config, users=users, keep_alive=keep_alive)

        async with self._node_lock.writer_lock:
            info = await self._handle_grpc_request(
                method=self._client.Start,
                request=req,
                timeout=timeout,
            )
            await self.connect(
                info.node_version,
                info.core_version,
                [
                    asyncio.create_task(self._check_node_health()),
                    asyncio.create_task(self._sync_user()),
                    asyncio.create_task(self._fetch_logs()),
                ],
            )
            self._metadata = {"authorization": f"Bearer {info.session_id}"}
            return info

    async def stop(self, timeout: int = 10) -> None:
        """Stop the node"""
        if await self.get_health() is Health.NOT_CONNECTED:
            return

        async with self._node_lock.writer_lock:
            await self.disconnect()

            await self._handle_grpc_request(
                method=self._client.Stop,
                request=service.Empty(),
                timeout=timeout,
            )

            self._metadata = None

    async def info(self, timeout: int = 10) -> service.BaseInfoResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetBaseInfo,
                request=service.Empty(),
                timeout=timeout,
            )

    async def get_system_stats(self, timeout: int = 10) -> service.SystemStatsResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetSystemStats,
                request=service.Empty(),
                timeout=timeout,
            )

    async def get_backend_stats(self, timeout: int = 10) -> service.BackendStatsResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetBackendStats,
                request=service.Empty(),
                timeout=timeout,
            )

    async def get_outbounds_stats(self, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetOutboundsStats,
                request=service.StatRequest(reset=reset),
                timeout=timeout,
            )

    async def get_outbound_stats(self, tag: str, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetOutboundStats,
                request=service.StatRequest(name=tag, reset=reset),
                timeout=timeout,
            )

    async def get_inbounds_stats(self, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetInboundsStats,
                request=service.StatRequest(reset=reset),
                timeout=timeout,
            )

    async def get_inbound_stats(self, tag: str, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetInboundStats,
                request=service.StatRequest(name=tag, reset=reset),
                timeout=timeout,
            )

    async def get_users_stats(self, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetUsersStats,
                request=service.StatRequest(reset=reset),
                timeout=timeout,
            )

    async def get_user_stats(self, email: str, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetUserStats,
                request=service.StatRequest(name=email, reset=reset),
                timeout=timeout,
            )

    async def user_online_stats(self, email: str, timeout: int = 10) -> service.OnlineStatResponse | None:
        await self.connected()

        async with self._node_lock.reader_lock:
            return await self._handle_grpc_request(
                method=self._client.GetUserOnlineStats,
                request=service.StatRequest(name=email),
                timeout=timeout,
            )

    async def sync_users(
        self, users: list[service.User], flush_queue: bool = False, timeout: int = 10
    ) -> service.Empty | None:
        await self.connected()
        if flush_queue:
            self.flush_user_queue()

        async with self._node_lock.writer_lock:
            return await self._handle_grpc_request(
                method=self._client.SyncUsers,
                request=service.Users(users=users),
                timeout=timeout,
            )

    async def _check_node_health(self):
        while True:
            last_health = await self.get_health()

            try:
                await self.get_backend_stats()
                if last_health != Health.HEALTHY:
                    await self.set_health(Health.HEALTHY)
            except Exception:
                if last_health != Health.BROKEN:
                    await self.set_health(Health.BROKEN)

            await asyncio.sleep(2)

    async def _fetch_logs(self):
        while True:
            health = await self.get_health()
            if health == Health.BROKEN:
                await asyncio.sleep(5)
                continue
            elif health == Health.NOT_CONNECTED:
                return

            try:
                async with self._client.GetLogs.open(metadata=self._metadata) as stream:
                    await stream.send_message(service.Empty())
                    while True:
                        log = await stream.recv_message()
                        if log is None:
                            continue
                        await self._logs_queue.put(log.detail)

            except Exception:
                await asyncio.sleep(3)
                continue

    async def _sync_user(self):
        while True:
            health = await self.get_health()

            if health == Health.BROKEN:
                await asyncio.sleep(5)
                continue
            elif health == Health.NOT_CONNECTED:
                return

            async with self._client.SyncUser.open(metadata=self._metadata) as stream:
                async with self._lock.reader_lock:
                    if self._user_queue is None or self._notify_queue is None:
                        return

                    user_task = asyncio.create_task(self._user_queue.get())
                    notify_task = asyncio.create_task(self._notify_queue.get())

                try:
                    # Wait for first completed operation
                    done, pending = await asyncio.wait([user_task, notify_task], return_when=asyncio.FIRST_COMPLETED)

                except asyncio.CancelledError:
                    # Cleanup if task is cancelled
                    user_task.cancel()
                    notify_task.cancel()
                    raise

                for task in pending:
                    task.cancel()

                if notify_task in done:
                    continue

                # Handle UserChan message
                if user_task in done:
                    user = user_task.result()
                    if user is None:
                        continue

                    try:
                        # Send the user through the gRPC stream
                        await stream.send_message(user)
                    except Exception:
                        await asyncio.sleep(5)
                        continue



from abc import ABC, abstractmethod

from GozargahNodeBridge.common import service_pb2 as service
from GozargahNodeBridge.controller import Controller


class GozargahNode(Controller, ABC):
    @abstractmethod
    async def start(
        self, config: str, backend_type: service.BackendType, users: list[service.User], keep_alive: int, timeout: int
    ) -> service.BaseInfoResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def stop(self, timeout: int) -> None:
        raise NotImplementedError

    @abstractmethod
    async def info(self, timeout: int) -> service.BaseInfoResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_system_stats(self, timeout: int) -> service.SystemStatsResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_backend_stats(self, timeout: int) -> service.BackendStatsResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_outbounds_stats(self, reset: bool, timeout: int) -> service.StatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_outbound_stats(self, tag: str, reset: bool, timeout: int) -> service.StatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_inbounds_stats(self, reset: bool, timeout: int) -> service.StatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_inbound_stats(self, tag: str, reset: bool, timeout: int) -> service.StatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_users_stats(self, reset: bool, timeout: int) -> service.StatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def get_user_stats(self, email: str, reset: bool, timeout: int) -> service.StatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def user_online_stats(self, email: str, timeout: int) -> service.OnlineStatResponse | None:
        raise NotImplementedError

    @abstractmethod
    async def sync_users(self, users: list[service.User], timeout: int) -> service.Empty | None:
        raise NotImplementedError

    @abstractmethod
    async def _check_node_health(self):
        raise NotImplementedError

    @abstractmethod
    async def _fetch_logs(self):
        raise NotImplementedError

    @abstractmethod
    async def _sync_user(self):
        raise NotImplementedError



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class BackendType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    XRAY: _ClassVar[BackendType]
XRAY: BackendType

class Empty(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class BaseInfoResponse(_message.Message):
    __slots__ = ("started", "core_version", "node_version", "session_id", "extra")
    STARTED_FIELD_NUMBER: _ClassVar[int]
    CORE_VERSION_FIELD_NUMBER: _ClassVar[int]
    NODE_VERSION_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    EXTRA_FIELD_NUMBER: _ClassVar[int]
    started: bool
    core_version: str
    node_version: str
    session_id: str
    extra: str
    def __init__(self, started: bool = ..., core_version: _Optional[str] = ..., node_version: _Optional[str] = ..., session_id: _Optional[str] = ..., extra: _Optional[str] = ...) -> None: ...

class Backend(_message.Message):
    __slots__ = ("type", "config", "users", "keep_alive")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    USERS_FIELD_NUMBER: _ClassVar[int]
    KEEP_ALIVE_FIELD_NUMBER: _ClassVar[int]
    type: BackendType
    config: str
    users: _containers.RepeatedCompositeFieldContainer[User]
    keep_alive: int
    def __init__(self, type: _Optional[_Union[BackendType, str]] = ..., config: _Optional[str] = ..., users: _Optional[_Iterable[_Union[User, _Mapping]]] = ..., keep_alive: _Optional[int] = ...) -> None: ...

class Log(_message.Message):
    __slots__ = ("detail",)
    DETAIL_FIELD_NUMBER: _ClassVar[int]
    detail: str
    def __init__(self, detail: _Optional[str] = ...) -> None: ...

class Stat(_message.Message):
    __slots__ = ("name", "type", "link", "value")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    LINK_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: str
    link: str
    value: int
    def __init__(self, name: _Optional[str] = ..., type: _Optional[str] = ..., link: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...

class StatResponse(_message.Message):
    __slots__ = ("stats",)
    STATS_FIELD_NUMBER: _ClassVar[int]
    stats: _containers.RepeatedCompositeFieldContainer[Stat]
    def __init__(self, stats: _Optional[_Iterable[_Union[Stat, _Mapping]]] = ...) -> None: ...

class StatRequest(_message.Message):
    __slots__ = ("name", "reset")
    NAME_FIELD_NUMBER: _ClassVar[int]
    RESET_FIELD_NUMBER: _ClassVar[int]
    name: str
    reset: bool
    def __init__(self, name: _Optional[str] = ..., reset: bool = ...) -> None: ...

class OnlineStatResponse(_message.Message):
    __slots__ = ("name", "ips")
    class IpsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    IPS_FIELD_NUMBER: _ClassVar[int]
    name: str
    ips: _containers.ScalarMap[str, int]
    def __init__(self, name: _Optional[str] = ..., ips: _Optional[_Mapping[str, int]] = ...) -> None: ...

class BackendStatsResponse(_message.Message):
    __slots__ = ("num_goroutine", "num_gc", "alloc", "total_alloc", "sys", "mallocs", "frees", "live_objects", "pause_total_ns", "uptime")
    NUM_GOROUTINE_FIELD_NUMBER: _ClassVar[int]
    NUM_GC_FIELD_NUMBER: _ClassVar[int]
    ALLOC_FIELD_NUMBER: _ClassVar[int]
    TOTAL_ALLOC_FIELD_NUMBER: _ClassVar[int]
    SYS_FIELD_NUMBER: _ClassVar[int]
    MALLOCS_FIELD_NUMBER: _ClassVar[int]
    FREES_FIELD_NUMBER: _ClassVar[int]
    LIVE_OBJECTS_FIELD_NUMBER: _ClassVar[int]
    PAUSE_TOTAL_NS_FIELD_NUMBER: _ClassVar[int]
    UPTIME_FIELD_NUMBER: _ClassVar[int]
    num_goroutine: int
    num_gc: int
    alloc: int
    total_alloc: int
    sys: int
    mallocs: int
    frees: int
    live_objects: int
    pause_total_ns: int
    uptime: int
    def __init__(self, num_goroutine: _Optional[int] = ..., num_gc: _Optional[int] = ..., alloc: _Optional[int] = ..., total_alloc: _Optional[int] = ..., sys: _Optional[int] = ..., mallocs: _Optional[int] = ..., frees: _Optional[int] = ..., live_objects: _Optional[int] = ..., pause_total_ns: _Optional[int] = ..., uptime: _Optional[int] = ...) -> None: ...

class SystemStatsResponse(_message.Message):
    __slots__ = ("mem_total", "mem_used", "cpu_cores", "cpu_usage", "incoming_bandwidth_speed", "outgoing_bandwidth_speed")
    MEM_TOTAL_FIELD_NUMBER: _ClassVar[int]
    MEM_USED_FIELD_NUMBER: _ClassVar[int]
    CPU_CORES_FIELD_NUMBER: _ClassVar[int]
    CPU_USAGE_FIELD_NUMBER: _ClassVar[int]
    INCOMING_BANDWIDTH_SPEED_FIELD_NUMBER: _ClassVar[int]
    OUTGOING_BANDWIDTH_SPEED_FIELD_NUMBER: _ClassVar[int]
    mem_total: int
    mem_used: int
    cpu_cores: int
    cpu_usage: float
    incoming_bandwidth_speed: int
    outgoing_bandwidth_speed: int
    def __init__(self, mem_total: _Optional[int] = ..., mem_used: _Optional[int] = ..., cpu_cores: _Optional[int] = ..., cpu_usage: _Optional[float] = ..., incoming_bandwidth_speed: _Optional[int] = ..., outgoing_bandwidth_speed: _Optional[int] = ...) -> None: ...

class Vmess(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class Vless(_message.Message):
    __slots__ = ("id", "flow")
    ID_FIELD_NUMBER: _ClassVar[int]
    FLOW_FIELD_NUMBER: _ClassVar[int]
    id: str
    flow: str
    def __init__(self, id: _Optional[str] = ..., flow: _Optional[str] = ...) -> None: ...

class Trojan(_message.Message):
    __slots__ = ("password",)
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    password: str
    def __init__(self, password: _Optional[str] = ...) -> None: ...

class Shadowsocks(_message.Message):
    __slots__ = ("password", "method")
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    METHOD_FIELD_NUMBER: _ClassVar[int]
    password: str
    method: str
    def __init__(self, password: _Optional[str] = ..., method: _Optional[str] = ...) -> None: ...

class Proxy(_message.Message):
    __slots__ = ("vmess", "vless", "trojan", "shadowsocks")
    VMESS_FIELD_NUMBER: _ClassVar[int]
    VLESS_FIELD_NUMBER: _ClassVar[int]
    TROJAN_FIELD_NUMBER: _ClassVar[int]
    SHADOWSOCKS_FIELD_NUMBER: _ClassVar[int]
    vmess: Vmess
    vless: Vless
    trojan: Trojan
    shadowsocks: Shadowsocks
    def __init__(self, vmess: _Optional[_Union[Vmess, _Mapping]] = ..., vless: _Optional[_Union[Vless, _Mapping]] = ..., trojan: _Optional[_Union[Trojan, _Mapping]] = ..., shadowsocks: _Optional[_Union[Shadowsocks, _Mapping]] = ...) -> None: ...

class User(_message.Message):
    __slots__ = ("email", "proxies", "inbounds")
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    PROXIES_FIELD_NUMBER: _ClassVar[int]
    INBOUNDS_FIELD_NUMBER: _ClassVar[int]
    email: str
    proxies: Proxy
    inbounds: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, email: _Optional[str] = ..., proxies: _Optional[_Union[Proxy, _Mapping]] = ..., inbounds: _Optional[_Iterable[str]] = ...) -> None: ...

class Users(_message.Message):
    __slots__ = ("users",)
    USERS_FIELD_NUMBER: _ClassVar[int]
    users: _containers.RepeatedCompositeFieldContainer[User]
    def __init__(self, users: _Optional[_Iterable[_Union[User, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: GozargahNodeBridge/common/service.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'GozargahNodeBridge/common/service.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'GozargahNodeBridge/common/service.proto\x12\x07service\"\x07\n\x05\x45mpty\"r\n\x10\x42\x61seInfoResponse\x12\x0f\n\x07started\x18\x01 \x01(\x08\x12\x14\n\x0c\x63ore_version\x18\x02 \x01(\t\x12\x14\n\x0cnode_version\x18\x03 \x01(\t\x12\x12\n\nsession_id\x18\x04 \x01(\t\x12\r\n\x05\x65xtra\x18\x05 \x01(\t\"o\n\x07\x42\x61\x63kend\x12\"\n\x04type\x18\x01 \x01(\x0e\x32\x14.service.BackendType\x12\x0e\n\x06\x63onfig\x18\x02 \x01(\t\x12\x1c\n\x05users\x18\x03 \x03(\x0b\x32\r.service.User\x12\x12\n\nkeep_alive\x18\x04 \x01(\x04\"\x15\n\x03Log\x12\x0e\n\x06\x64\x65tail\x18\x01 \x01(\t\"?\n\x04Stat\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0c\n\x04link\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\x03\",\n\x0cStatResponse\x12\x1c\n\x05stats\x18\x01 \x03(\x0b\x32\r.service.Stat\"*\n\x0bStatRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05reset\x18\x02 \x01(\x08\"\x81\x01\n\x12OnlineStatResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x31\n\x03ips\x18\x02 \x03(\x0b\x32$.service.OnlineStatResponse.IpsEntry\x1a*\n\x08IpsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"\xcc\x01\n\x14\x42\x61\x63kendStatsResponse\x12\x15\n\rnum_goroutine\x18\x01 \x01(\r\x12\x0e\n\x06num_gc\x18\x02 \x01(\r\x12\r\n\x05\x61lloc\x18\x03 \x01(\x04\x12\x13\n\x0btotal_alloc\x18\x04 \x01(\x04\x12\x0b\n\x03sys\x18\x05 \x01(\x04\x12\x0f\n\x07mallocs\x18\x06 \x01(\x04\x12\r\n\x05\x66rees\x18\x07 \x01(\x04\x12\x14\n\x0clive_objects\x18\x08 \x01(\x04\x12\x16\n\x0epause_total_ns\x18\t \x01(\x04\x12\x0e\n\x06uptime\x18\n \x01(\r\"\xa4\x01\n\x13SystemStatsResponse\x12\x11\n\tmem_total\x18\x01 \x01(\x04\x12\x10\n\x08mem_used\x18\x02 \x01(\x04\x12\x11\n\tcpu_cores\x18\x03 \x01(\x04\x12\x11\n\tcpu_usage\x18\x04 \x01(\x01\x12 \n\x18incoming_bandwidth_speed\x18\x05 \x01(\x04\x12 \n\x18outgoing_bandwidth_speed\x18\x06 \x01(\x04\"\x13\n\x05Vmess\x12\n\n\x02id\x18\x01 \x01(\t\"!\n\x05Vless\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04\x66low\x18\x02 \x01(\t\"\x1a\n\x06Trojan\x12\x10\n\x08password\x18\x01 \x01(\t\"/\n\x0bShadowsocks\x12\x10\n\x08password\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\"\x91\x01\n\x05Proxy\x12\x1d\n\x05vmess\x18\x01 \x01(\x0b\x32\x0e.service.Vmess\x12\x1d\n\x05vless\x18\x02 \x01(\x0b\x32\x0e.service.Vless\x12\x1f\n\x06trojan\x18\x03 \x01(\x0b\x32\x0f.service.Trojan\x12)\n\x0bshadowsocks\x18\x04 \x01(\x0b\x32\x14.service.Shadowsocks\"H\n\x04User\x12\r\n\x05\x65mail\x18\x01 \x01(\t\x12\x1f\n\x07proxies\x18\x02 \x01(\x0b\x32\x0e.service.Proxy\x12\x10\n\x08inbounds\x18\x03 \x03(\t\"%\n\x05Users\x12\x1c\n\x05users\x18\x01 \x03(\x0b\x32\r.service.User*\x17\n\x0b\x42\x61\x63kendType\x12\x08\n\x04XRAY\x10\x00\x32\x92\x07\n\x0bNodeService\x12\x36\n\x05Start\x12\x10.service.Backend\x1a\x19.service.BaseInfoResponse\"\x00\x12(\n\x04Stop\x12\x0e.service.Empty\x1a\x0e.service.Empty\"\x00\x12:\n\x0bGetBaseInfo\x12\x0e.service.Empty\x1a\x19.service.BaseInfoResponse\"\x00\x12+\n\x07GetLogs\x12\x0e.service.Empty\x1a\x0c.service.Log\"\x00\x30\x01\x12@\n\x0eGetSystemStats\x12\x0e.service.Empty\x1a\x1c.service.SystemStatsResponse\"\x00\x12\x42\n\x0fGetBackendStats\x12\x0e.service.Empty\x1a\x1d.service.BackendStatsResponse\"\x00\x12\x42\n\x11GetOutboundsStats\x12\x14.service.StatRequest\x1a\x15.service.StatResponse\"\x00\x12\x41\n\x10GetOutboundStats\x12\x14.service.StatRequest\x1a\x15.service.StatResponse\"\x00\x12\x41\n\x10GetInboundsStats\x12\x14.service.StatRequest\x1a\x15.service.StatResponse\"\x00\x12@\n\x0fGetInboundStats\x12\x14.service.StatRequest\x1a\x15.service.StatResponse\"\x00\x12>\n\rGetUsersStats\x12\x14.service.StatRequest\x1a\x15.service.StatResponse\"\x00\x12=\n\x0cGetUserStats\x12\x14.service.StatRequest\x1a\x15.service.StatResponse\"\x00\x12I\n\x12GetUserOnlineStats\x12\x14.service.StatRequest\x1a\x1b.service.OnlineStatResponse\"\x00\x12-\n\x08SyncUser\x12\r.service.User\x1a\x0e.service.Empty\"\x00(\x01\x12-\n\tSyncUsers\x12\x0e.service.Users\x1a\x0e.service.Empty\"\x00\x42\x18Z\x16marzban_node_go/commonb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'GozargahNodeBridge.common.service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\026marzban_node_go/common'
  _globals['_ONLINESTATRESPONSE_IPSENTRY']._loaded_options = None
  _globals['_ONLINESTATRESPONSE_IPSENTRY']._serialized_options = b'8\001'
  _globals['_BACKENDTYPE']._serialized_start=1368
  _globals['_BACKENDTYPE']._serialized_end=1391
  _globals['_EMPTY']._serialized_start=52
  _globals['_EMPTY']._serialized_end=59
  _globals['_BASEINFORESPONSE']._serialized_start=61
  _globals['_BASEINFORESPONSE']._serialized_end=175
  _globals['_BACKEND']._serialized_start=177
  _globals['_BACKEND']._serialized_end=288
  _globals['_LOG']._serialized_start=290
  _globals['_LOG']._serialized_end=311
  _globals['_STAT']._serialized_start=313
  _globals['_STAT']._serialized_end=376
  _globals['_STATRESPONSE']._serialized_start=378
  _globals['_STATRESPONSE']._serialized_end=422
  _globals['_STATREQUEST']._serialized_start=424
  _globals['_STATREQUEST']._serialized_end=466
  _globals['_ONLINESTATRESPONSE']._serialized_start=469
  _globals['_ONLINESTATRESPONSE']._serialized_end=598
  _globals['_ONLINESTATRESPONSE_IPSENTRY']._serialized_start=556
  _globals['_ONLINESTATRESPONSE_IPSENTRY']._serialized_end=598
  _globals['_BACKENDSTATSRESPONSE']._serialized_start=601
  _globals['_BACKENDSTATSRESPONSE']._serialized_end=805
  _globals['_SYSTEMSTATSRESPONSE']._serialized_start=808
  _globals['_SYSTEMSTATSRESPONSE']._serialized_end=972
  _globals['_VMESS']._serialized_start=974
  _globals['_VMESS']._serialized_end=993
  _globals['_VLESS']._serialized_start=995
  _globals['_VLESS']._serialized_end=1028
  _globals['_TROJAN']._serialized_start=1030
  _globals['_TROJAN']._serialized_end=1056
  _globals['_SHADOWSOCKS']._serialized_start=1058
  _globals['_SHADOWSOCKS']._serialized_end=1105
  _globals['_PROXY']._serialized_start=1108
  _globals['_PROXY']._serialized_end=1253
  _globals['_USER']._serialized_start=1255
  _globals['_USER']._serialized_end=1327
  _globals['_USERS']._serialized_start=1329
  _globals['_USERS']._serialized_end=1366
  _globals['_NODESERVICE']._serialized_start=1394
  _globals['_NODESERVICE']._serialized_end=2308
# @@protoc_insertion_point(module_scope)



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: GozargahNodeBridge/common/service.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

import GozargahNodeBridge.common.service_pb2


class NodeServiceBase(abc.ABC):

    @abc.abstractmethod
    async def Start(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Backend, GozargahNodeBridge.common.service_pb2.BaseInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Stop(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Empty, GozargahNodeBridge.common.service_pb2.Empty]') -> None:
        pass

    @abc.abstractmethod
    async def GetBaseInfo(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Empty, GozargahNodeBridge.common.service_pb2.BaseInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetLogs(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Empty, GozargahNodeBridge.common.service_pb2.Log]') -> None:
        pass

    @abc.abstractmethod
    async def GetSystemStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Empty, GozargahNodeBridge.common.service_pb2.SystemStatsResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetBackendStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Empty, GozargahNodeBridge.common.service_pb2.BackendStatsResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetOutboundsStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.StatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetOutboundStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.StatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetInboundsStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.StatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetInboundStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.StatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetUsersStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.StatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetUserStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.StatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def GetUserOnlineStats(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.StatRequest, GozargahNodeBridge.common.service_pb2.OnlineStatResponse]') -> None:
        pass

    @abc.abstractmethod
    async def SyncUser(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.User, GozargahNodeBridge.common.service_pb2.Empty]') -> None:
        pass

    @abc.abstractmethod
    async def SyncUsers(self, stream: 'grpclib.server.Stream[GozargahNodeBridge.common.service_pb2.Users, GozargahNodeBridge.common.service_pb2.Empty]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/service.NodeService/Start': grpclib.const.Handler(
                self.Start,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.Backend,
                GozargahNodeBridge.common.service_pb2.BaseInfoResponse,
            ),
            '/service.NodeService/Stop': grpclib.const.Handler(
                self.Stop,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.Empty,
                GozargahNodeBridge.common.service_pb2.Empty,
            ),
            '/service.NodeService/GetBaseInfo': grpclib.const.Handler(
                self.GetBaseInfo,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.Empty,
                GozargahNodeBridge.common.service_pb2.BaseInfoResponse,
            ),
            '/service.NodeService/GetLogs': grpclib.const.Handler(
                self.GetLogs,
                grpclib.const.Cardinality.UNARY_STREAM,
                GozargahNodeBridge.common.service_pb2.Empty,
                GozargahNodeBridge.common.service_pb2.Log,
            ),
            '/service.NodeService/GetSystemStats': grpclib.const.Handler(
                self.GetSystemStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.Empty,
                GozargahNodeBridge.common.service_pb2.SystemStatsResponse,
            ),
            '/service.NodeService/GetBackendStats': grpclib.const.Handler(
                self.GetBackendStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.Empty,
                GozargahNodeBridge.common.service_pb2.BackendStatsResponse,
            ),
            '/service.NodeService/GetOutboundsStats': grpclib.const.Handler(
                self.GetOutboundsStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.StatResponse,
            ),
            '/service.NodeService/GetOutboundStats': grpclib.const.Handler(
                self.GetOutboundStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.StatResponse,
            ),
            '/service.NodeService/GetInboundsStats': grpclib.const.Handler(
                self.GetInboundsStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.StatResponse,
            ),
            '/service.NodeService/GetInboundStats': grpclib.const.Handler(
                self.GetInboundStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.StatResponse,
            ),
            '/service.NodeService/GetUsersStats': grpclib.const.Handler(
                self.GetUsersStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.StatResponse,
            ),
            '/service.NodeService/GetUserStats': grpclib.const.Handler(
                self.GetUserStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.StatResponse,
            ),
            '/service.NodeService/GetUserOnlineStats': grpclib.const.Handler(
                self.GetUserOnlineStats,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.StatRequest,
                GozargahNodeBridge.common.service_pb2.OnlineStatResponse,
            ),
            '/service.NodeService/SyncUser': grpclib.const.Handler(
                self.SyncUser,
                grpclib.const.Cardinality.STREAM_UNARY,
                GozargahNodeBridge.common.service_pb2.User,
                GozargahNodeBridge.common.service_pb2.Empty,
            ),
            '/service.NodeService/SyncUsers': grpclib.const.Handler(
                self.SyncUsers,
                grpclib.const.Cardinality.UNARY_UNARY,
                GozargahNodeBridge.common.service_pb2.Users,
                GozargahNodeBridge.common.service_pb2.Empty,
            ),
        }


class NodeServiceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.Start = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/Start',
            GozargahNodeBridge.common.service_pb2.Backend,
            GozargahNodeBridge.common.service_pb2.BaseInfoResponse,
        )
        self.Stop = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/Stop',
            GozargahNodeBridge.common.service_pb2.Empty,
            GozargahNodeBridge.common.service_pb2.Empty,
        )
        self.GetBaseInfo = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetBaseInfo',
            GozargahNodeBridge.common.service_pb2.Empty,
            GozargahNodeBridge.common.service_pb2.BaseInfoResponse,
        )
        self.GetLogs = grpclib.client.UnaryStreamMethod(
            channel,
            '/service.NodeService/GetLogs',
            GozargahNodeBridge.common.service_pb2.Empty,
            GozargahNodeBridge.common.service_pb2.Log,
        )
        self.GetSystemStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetSystemStats',
            GozargahNodeBridge.common.service_pb2.Empty,
            GozargahNodeBridge.common.service_pb2.SystemStatsResponse,
        )
        self.GetBackendStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetBackendStats',
            GozargahNodeBridge.common.service_pb2.Empty,
            GozargahNodeBridge.common.service_pb2.BackendStatsResponse,
        )
        self.GetOutboundsStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetOutboundsStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.StatResponse,
        )
        self.GetOutboundStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetOutboundStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.StatResponse,
        )
        self.GetInboundsStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetInboundsStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.StatResponse,
        )
        self.GetInboundStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetInboundStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.StatResponse,
        )
        self.GetUsersStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetUsersStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.StatResponse,
        )
        self.GetUserStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetUserStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.StatResponse,
        )
        self.GetUserOnlineStats = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/GetUserOnlineStats',
            GozargahNodeBridge.common.service_pb2.StatRequest,
            GozargahNodeBridge.common.service_pb2.OnlineStatResponse,
        )
        self.SyncUser = grpclib.client.StreamUnaryMethod(
            channel,
            '/service.NodeService/SyncUser',
            GozargahNodeBridge.common.service_pb2.User,
            GozargahNodeBridge.common.service_pb2.Empty,
        )
        self.SyncUsers = grpclib.client.UnaryUnaryMethod(
            channel,
            '/service.NodeService/SyncUsers',
            GozargahNodeBridge.common.service_pb2.Users,
            GozargahNodeBridge.common.service_pb2.Empty,
        )



import asyncio

import GozargahNodeBridge as Bridge

address = "172.27.158.135"
port = 62050
client_cert_file = "certs/ssl_client_cert.pem"
client_key_file = "certs/ssl_client_key.pem"
server_ca_file = "certs/ssl_cert.pem"
config_file = "config/xray.json"

with open(config_file, "r") as f:
    config = f.read()

with open(client_cert_file, "r") as f:
    client_cert_content = f.read()

with open(client_key_file, "r") as f:
    client_key_content = f.read()

with open(server_ca_file, "r") as f:
    server_ca_content = f.read()


async def main():
    node = Bridge.create_node(
        connection=Bridge.NodeType.grpc,
        address=address,
        port=port,
        client_cert=client_cert_content,
        client_key=client_key_content,
        server_ca=server_ca_content,
        max_logs=100,
        extra={"id": 1},
    )

    await node.start(config=config, backend_type=0, users=[], timeout=20)

    user = Bridge.create_user(
        email="jeff", proxies=Bridge.create_proxy(vmess_id="0d59268a-9847-4218-ae09-65308eb52e08"), inbounds=[]
    )

    await node.update_user(user)

    await asyncio.sleep(5)

    stats = await node.get_outbounds_stats()
    print(stats)

    logs = await node.get_logs()

    print(await logs.get())

    await node.stop()


asyncio.run(main())



import ssl
import asyncio
from enum import IntEnum
from pathlib import Path

from aiorwlock import RWLock

from GozargahNodeBridge.common.service_pb2 import User
from GozargahNodeBridge.utils import string_to_temp_file


class RollingQueue(asyncio.Queue):
    async def put(self, item):
        while self.maxsize > 0 and self.full():
            await self.get()
        await super().put(item)


class NodeAPIError(Exception):
    def __init__(self, code, detail):
        self.code = code
        self.detail = detail

    def __str__(self):
        return f"NodeAPIError(code={self.code}, detail={self.detail})"


class Health(IntEnum):
    NOT_CONNECTED = 0
    BROKEN = 1
    HEALTHY = 2
    INVALID = 3


class Controller:
    def __init__(
        self, client_cert: str, client_key: str, server_ca: str, extra: dict | None = None, max_logs: int = 1000
    ):
        self.max_logs = max_logs
        self._temp_files = []
        if extra is None:
            extra = {}
        try:
            ca_path = string_to_temp_file(server_ca)
            cert_path = string_to_temp_file(client_cert)
            key_path = string_to_temp_file(client_key)
            self._temp_files.extend([Path(ca_path), Path(cert_path), Path(key_path)])

            self.ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            self.ctx.set_alpn_protocols(["h2"])
            self.ctx.load_verify_locations(cafile=ca_path)
            self.ctx.load_cert_chain(certfile=cert_path, keyfile=key_path)
            self.ctx.check_hostname = False
            self.ctx.verify_mode = ssl.CERT_REQUIRED

        except (ssl.SSLError, IOError) as e:
            self._cleanup_temp_files()
            raise NodeAPIError(-1, f"SSL initialization failed: {str(e)}")

        self._health = Health.NOT_CONNECTED
        self._user_queue: asyncio.Queue[User] | None = None
        self._logs_queue: RollingQueue[str] | None = None
        self._notify_queue: asyncio.Queue | None = None
        self._tasks: list[asyncio.Task] = []
        self._node_version = ""
        self._core_version = ""
        self._extra = extra
        self._lock = RWLock()

    def _cleanup_temp_files(self):
        """Clean up temporary certificate files"""
        for path in self._temp_files:
            try:
                path.unlink(missing_ok=True)
            except Exception:
                pass

    def __del__(self):
        self._cleanup_temp_files()

    async def set_health(self, health: Health):
        async with self._lock.writer_lock:
            if self._health is Health.INVALID:
                return
            if health == Health.BROKEN and self._health != Health.BROKEN:
                if self._notify_queue:
                    await self._notify_queue.put(None)
            self._health = health

    async def get_health(self) -> Health:
        async with self._lock.reader_lock:
            return self._health

    async def connected(self) -> bool:
        health = await self.get_health()
        if health is Health.NOT_CONNECTED:
            raise NodeAPIError(code=-3, detail="Node is not connected")
        elif health is Health.INVALID:
            raise NodeAPIError(code=-4, detail="Invalid node")
        return True

    async def update_user(self, user: User):
        await self.connected()
        async with self._lock.reader_lock:
            if self._user_queue:
                await self._user_queue.put(user)

    async def flush_user_queue(self):
        await self.connected()
        async with self._lock.writer_lock:
            self._user_queue.empty()

    async def get_logs(self) -> asyncio.Queue | None:
        await self.connected()
        async with self._lock.reader_lock:
            return self._logs_queue

    async def flush_logs_queue(self):
        await self.connected()
        async with self._lock.writer_lock:
            self._logs_queue.empty()

    @property
    async def node_version(self) -> str:
        async with self._lock.reader_lock:
            return self._node_version

    @property
    async def core_version(self) -> str:
        async with self._lock.reader_lock:
            return self._core_version

    async def get_extra(self) -> dict:
        async with self._lock.reader_lock:
            return self._extra

    async def connect(self, node_version: str, core_version: str, tasks: list[asyncio.Task] | None = None):
        if tasks is None:
            tasks = []
        async with self._lock.writer_lock:
            self._user_queue = asyncio.Queue()
            self._notify_queue = asyncio.Queue()
            self._logs_queue = RollingQueue(self.max_logs)
            self._tasks = tasks
            self._node_version = node_version
            self._core_version = core_version
            self._health = Health.HEALTHY

    async def disconnect(self):
        await self.set_health(Health.NOT_CONNECTED)

        async with self._lock.writer_lock:
            for task in self._tasks:
                task.cancel()

            await asyncio.gather(*self._tasks, return_exceptions=True)
            self._tasks.clear()

            if self._user_queue:
                await self._user_queue.put(None)
                self._user_queue = None

            if self._notify_queue:
                await self._notify_queue.put(None)
                self._notify_queue = None

            if self._logs_queue:
                await self._logs_queue.put(None)
                self._logs_queue = None

            self._node_version = ""
            self._core_version = ""



"""
Gozargah Node Bridge
A Python library for connecting to Gozargah nodes.
"""

__version__ = "0.0.20"
__author__ = "M03ED"


from enum import Enum

from GozargahNodeBridge.abstract_node import GozargahNode
from GozargahNodeBridge.grpclib import Node as GrpcNode
from GozargahNodeBridge.rest import Node as RestNode
from GozargahNodeBridge.controller import NodeAPIError, Health
from GozargahNodeBridge.utils import create_user, create_proxy


class NodeType(str, Enum):
    grpc = "grpc"
    rest = "rest"


def create_node(
    connection: NodeType,
    address: str,
    port: int,
    client_cert: str,
    client_key: str,
    server_ca: str,
    max_logs: int = 1000,
    extra: dict = {},
) -> GozargahNode:
    """
    Create and initialize a node instance.

    Args:
         connection (NodeType): The type of connection to use for the node.
             Must be `NodeType.GRPC` or `NodeType.REST`.
         address (str): The address of the node.
             This must be a valid IPv4, IPv6 or valid domain.
         port (int): The port number for the node connection.
         client_cert (str): The SSL certificate as a string (not a file path).
         client_key (str): The SSL private key as a string (not a file path).
         server_ca (str): The server SSL certificate as a string (not a file path).
         extra (dict): extra data you need in production.

    Returns:
        Node | None: A `Node` instance if successfully created; otherwise, `None`.

    Raises:
        NodeAPIError: If the `connection` failed.

    Note:
        - The `address` must be a valid IP address or domain.
        - The `client_cert`, `client_key` and `server_ca` should be the content of the SSL certificate
          and private key as strings, not file paths.
    """

    if connection is NodeType.grpc:
        node = GrpcNode(
            address=address,
            port=port,
            client_cert=client_cert,
            client_key=client_key,
            server_ca=server_ca,
            extra=extra,
            max_logs=max_logs,
        )

    elif connection is NodeType.rest:
        node = RestNode(
            address=address,
            port=port,
            client_cert=client_cert,
            client_key=client_key,
            server_ca=server_ca,
            extra=extra,
            max_logs=max_logs,
        )

    else:
        raise ValueError("invalid backend type")

    return node


__all__ = [
    "NodeType",
    "Node",
    "NodeAPIError",
    "Health",
    "create_user",
    "create_proxy",
    "create_node",
]



import asyncio
import httpx

from aiorwlock import RWLock
from google.protobuf.message import Message, DecodeError
from GozargahNodeBridge.controller import NodeAPIError, Health
from GozargahNodeBridge.common import service_pb2 as service
from GozargahNodeBridge.abstract_node import GozargahNode


class Node(GozargahNode):
    def __init__(
        self,
        address: str,
        port: int,
        client_cert: str,
        client_key: str,
        server_ca: str,
        extra: dict | None = None,
        max_logs: int = 1000,
    ):
        super().__init__(client_cert, client_key, server_ca, extra, max_logs)

        url = f"https://{address.strip('/')}:{port}/"

        self._client = httpx.AsyncClient(
            http2=True,
            verify=self.ctx,
            headers={"Content-Type": "application/x-protobuf"},
            base_url=url,
            timeout=httpx.Timeout(None),
            limits=httpx.Limits(max_keepalive_connections=1),
        )

        self._node_lock = RWLock()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()
        await self._client.aclose()
        self._cleanup_temp_files()

    def __del__(self):
        self._cleanup_temp_files()

    def _serialize_protobuf(self, proto_message: Message) -> bytes:
        """Serialize a protobuf message to bytes."""
        return proto_message.SerializeToString()

    def _deserialize_protobuf(self, proto_class: type[Message], data: bytes) -> Message:
        """Deserialize bytes into a protobuf message."""
        proto_instance = proto_class()
        try:
            proto_instance.ParseFromString(data)
        except DecodeError as e:
            raise NodeAPIError(code=-2, detail=f"Error deserialising protobuf: {e}")
        return proto_instance

    def _handle_error(self, error: Exception):
        if isinstance(error, httpx.RemoteProtocolError):
            raise NodeAPIError(code=-1, detail=f"Server closed connection: {error}")
        elif isinstance(error, httpx.HTTPStatusError):
            raise NodeAPIError(code=error.response.status_code, detail=f"HTTP error: {error.response.text}")
        elif isinstance(error, httpx.ConnectError) or isinstance(error, httpx.ReadTimeout):
            raise NodeAPIError(code=-1, detail=f"Connection error: {error}")
        else:
            raise NodeAPIError(0, str(error))

    async def _make_request(
        self,
        method: str,
        endpoint: str,
        timeout: int,
        proto_message: Message = None,
        proto_response_class: type[Message] = None,
    ) -> Message:
        """Handle common REST API call logic with protobuf support (async)."""
        request_data = None

        if proto_message:
            request_data = self._serialize_protobuf(proto_message)

        try:
            response = await self._client.request(
                method=method,
                url=endpoint,
                content=request_data,
                timeout=timeout,
            )
            response.raise_for_status()

            if proto_response_class:
                return self._deserialize_protobuf(proto_response_class, response.content)
            return response.content

        except Exception as e:
            self._handle_error(e)

    async def start(
        self,
        config: str,
        backend_type: service.BackendType,
        users: list[service.User],
        keep_alive: int = 0,
        timeout: int = 15,
    ):
        health = await self.get_health()
        if health in (Health.BROKEN, Health.HEALTHY):
            await self.stop()
        elif health is Health.INVALID:
            raise NodeAPIError(code=-4, detail="Invalid node")

        async with self._node_lock.writer_lock:
            response = await self._make_request(
                method="POST",
                endpoint="start",
                timeout=timeout,
                proto_message=service.Backend(type=backend_type, config=config, users=users, keep_alive=keep_alive),
                proto_response_class=service.BaseInfoResponse,
            )

            self._client.headers["Authorization"] = f"Bearer {response.session_id}"
            await self.connect(
                response.node_version,
                response.core_version,
                [
                    asyncio.create_task(self._check_node_health()),
                    asyncio.create_task(self._sync_user()),
                    asyncio.create_task(self._fetch_logs()),
                ],
            )

        return response

    async def stop(self, timeout: int = 10) -> None:
        if await self.get_health() is Health.NOT_CONNECTED:
            return
        async with self._node_lock.writer_lock:
            await self.disconnect()
            await self._make_request(method="PUT", endpoint="stop", timeout=timeout)

    async def info(self, timeout: int = 10) -> service.BaseInfoResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(method="GET", endpoint="info", timeout=timeout)

    async def get_system_stats(self, timeout: int = 10) -> service.SystemStatsResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(method="GET", endpoint="stats/system", timeout=timeout)

    async def get_backend_stats(self, timeout: int = 10) -> service.BackendStatsResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(method="GET", endpoint="stats/backend", timeout=timeout)

    async def get_outbounds_stats(self, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/outbounds",
                timeout=timeout,
                proto_message=service.StatRequest(reset=reset),
                proto_response_class=service.StatResponse,
            )

    async def get_outbound_stats(self, tag: str, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/outbound",
                timeout=timeout,
                proto_message=service.StatRequest(reset=reset, name=tag),
                proto_response_class=service.StatResponse,
            )

    async def get_inbounds_stats(self, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/inbounds",
                timeout=timeout,
                proto_message=service.StatRequest(reset=reset),
                proto_response_class=service.StatResponse,
            )

    async def get_inbound_stats(self, tag: str, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/inbound",
                timeout=timeout,
                proto_message=service.StatRequest(reset=reset, name=tag),
                proto_response_class=service.StatResponse,
            )

    async def get_users_stats(self, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/users",
                timeout=timeout,
                proto_message=service.StatRequest(reset=reset),
                proto_response_class=service.StatResponse,
            )

    async def get_user_stats(self, email: str, reset: bool = True, timeout: int = 10) -> service.StatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/user",
                timeout=timeout,
                proto_message=service.StatRequest(reset=reset, name=email),
                proto_response_class=service.StatResponse,
            )

    async def user_online_stats(self, email: str, timeout: int = 10) -> service.OnlineStatResponse | None:
        await self.connected()
        async with self._node_lock.reader_lock:
            return await self._make_request(
                method="GET",
                endpoint="stats/user/online",
                timeout=timeout,
                proto_message=service.StatRequest(name=email),
                proto_response_class=service.OnlineStatResponse,
            )

    async def sync_users(
        self, users: list[service.User], flush_queue: bool = False, timeout: int = 10
    ) -> service.Empty | None:
        await self.connected()
        if flush_queue:
            self.flush_user_queue()

        async with self._node_lock.writer_lock:
            return await self._make_request(
                method="POST",
                endpoint="users/sync",
                timeout=timeout,
                proto_message=service.Users(users=users),
                proto_response_class=service.Empty,
            )

    async def _check_node_health(self):
        while True:
            last_health = await self.get_health()

            try:
                await self.get_backend_stats()
                if last_health != Health.HEALTHY:
                    await self.set_health(Health.HEALTHY)
            except Exception:
                if last_health != Health.BROKEN:
                    await self.set_health(Health.BROKEN)

            await asyncio.sleep(2)

    async def _fetch_logs(self):
        while True:
            health = await self.get_health()
            if health == Health.BROKEN:
                await asyncio.sleep(5)
                continue
            elif health == Health.NOT_CONNECTED:
                return

            try:
                async with self._client.stream("GET", "/logs", timeout=None) as response:
                    buffer = b""

                    async for chunk in response.aiter_bytes():
                        buffer += chunk

                        while b"\n" in buffer:
                            line, buffer = buffer.split(b"\n", 1)
                            line = line.decode().strip()

                            if line:
                                if self._logs_queue:
                                    await self._logs_queue.put(line)

            except Exception:
                await asyncio.sleep(3)
                continue

    async def _sync_user(self) -> None:
        while True:
            health = await self.get_health()

            if health == Health.BROKEN:
                await asyncio.sleep(5)
                continue
            elif health == Health.NOT_CONNECTED:
                return

            async with self._lock.reader_lock:
                if self._user_queue is None or self._notify_queue is None:
                    return

                user_task = asyncio.create_task(self._user_queue.get())
                notify_task = asyncio.create_task(self._notify_queue.get())

            try:
                # Wait for first completed operation
                done, pending = await asyncio.wait([user_task, notify_task], return_when=asyncio.FIRST_COMPLETED)

            except asyncio.CancelledError:
                # Cleanup if task is cancelled
                user_task.cancel()
                notify_task.cancel()
                raise

            # Cancel pending tasks to avoid leaks
            for task in pending:
                task.cancel()

            if notify_task in done:
                continue

            # Handle UserChan message
            if user_task in done:
                user = user_task.result()
                if user is None:
                    continue

                try:
                    await self._make_request(
                        method="PUT",
                        endpoint="user/sync",
                        timeout=10,
                        proto_message=user,
                        proto_response_class=service.Empty,
                    )
                except Exception:
                    await asyncio.sleep(5)
                    continue
